import { stockDatabase } from "./stockdb.clickhouse.js";
import { calendarDatabase } from "./calendardb.optiondb.lmdbx.js";
import type { CalendarKey } from "./calendardb.interfaces.js";
import type { Aggregate } from "./interfaces.js";
import { nextDate } from "./lib/util.js";

type BacktestInput = {
  symbol: string;
  startDate: string;
  endDate: string;
  /** Between 0 and 1. The frequency that similar calendars have historically ended (i.e. within the last hour) at a higher price than the current calendar's price. */
  historicalProbabilityOfSuccess?: number;
  initialAvailableValue?: number;
};
export async function backtest({
  symbol,
  startDate,
  endDate,
  historicalProbabilityOfSuccess = 0.8,
  initialAvailableValue: initialBuyingPower = 2000,
}: BacktestInput) {
  let buyingPower = initialBuyingPower;
  const portfolio = new Set<CalendarKey>();
  // for each day:
  for (
    let date = startDate, didBuyCalendar = false;
    date <= endDate;
    date = nextDate(date), didBuyCalendar = false
  ) {
    console.log("Current Date:", date);
    const calendars = await calendarDatabase.getCalendars({
      key: { symbol },
      date,
    });
    const stockAggregates = await stockDatabase.getAggregates({
      key: symbol,
      date,
    });
    const calendarsAggregates = new Map<
      CalendarKey,
      Array<Pick<Aggregate<CalendarKey>, "tsStart" | "open" | "close">>
    >();
    for (const calendar of calendars) {
      calendarsAggregates.set(
        calendar,
        await calendarDatabase.getAggregates({
          key: {
            ...calendar,
          },
          date,
        })
      );
    }
    // for each minute of that day for which we have a stock candlestick:
    for (const stockAggregate of stockAggregates) {
      // console.log("Current Time:", new Date(stockAggregate.tsStart));
      // filter-out calendars that are far-from-the-money (10%)
      const calendarsNearTheMoney = calendars.filter(
        ({ strike }) =>
          Math.abs((stockAggregate.open - strike) / stockAggregate.open) < 0.1
      );
      // for each relevant calendar on that day:
      for (const calendar of calendarsNearTheMoney) {
        const strikePercentageFromTheMoney = Math.abs(
          (stockAggregate.open - calendar.strike) / stockAggregate.open
        );
        /** In days. */
        const calendarSpan =
          (new Date(calendar.backExpirationDate).valueOf() -
            new Date(calendar.frontExpirationDate).valueOf()) /
          (1000 * 60 * 60 * 24);
        const targetCalendarPrice =
          await calendarDatabase.getTargetPriceByProbability({
            symbol,
            calendarSpan,
            strikePercentageFromTheMoney,
            historicalProbabilityOfSuccess,
          });
        const calendarAggregates = calendarsAggregates.get(calendar);
        const calendarAggregateAtCurrentTime = calendarAggregates.find(
          ({ tsStart }) => tsStart === stockAggregate.tsStart
        );
        // if there exists a matching calendar candlestick for the current minute:
        if (calendarAggregateAtCurrentTime) {
          // if the current candlestick is a good price (i.e. less than the target price):
          const minCalendarPriceInCandlestick = Math.min(
            calendarAggregateAtCurrentTime.open,
            calendarAggregateAtCurrentTime.close
          );
          if (
            minCalendarPriceInCandlestick < targetCalendarPrice &&
            minCalendarPriceInCandlestick >
              0.07 /* sometimes the calendar price is zero or negative, which is of course impossible; some institution got a good deal */
          ) {
            // if we can afford to buy the calendar:
            if (buyingPower > minCalendarPriceInCandlestick) {
              // buy the calendar, and continue to the next day:
              portfolio.add(calendar);
              buyingPower = buyingPower - minCalendarPriceInCandlestick * 100;
              console.log(
                "Bought",
                calendar,
                "for",
                minCalendarPriceInCandlestick * 100,
                "...$",
                buyingPower,
                "left"
              );
              didBuyCalendar = true;
            }
          }
        }
        if (didBuyCalendar) {
          break;
        }
      }
      if (didBuyCalendar) {
        break;
      }
    }

    // for each calendar in portfolio, if today is the last day, close the position:
    for (const calendar of portfolio.values()) {
      if (calendar.frontExpirationDate === date) {
        const calendarClosingPrice = await calendarDatabase.getClosingPrice({
          key: {
            ...calendar,
          },
        });
        portfolio.delete(calendar);
        buyingPower = buyingPower + calendarClosingPrice * 100;
        console.log(
          "Sold",
          calendar,
          "for",
          calendarClosingPrice,
          "...$",
          buyingPower,
          "left"
        );
      }
    }
  }

  console.log("Ending Buying Power:", buyingPower);
  console.log("Portfolio:", portfolio.values());
}
