import * as Polygon from "./polygon.js";
import sqlite3 from "sqlite3";
import { open } from "sqlite";
import { clickhouse, query } from "./clickhouse.js";
import { OptionContract } from "./polygon.js";
import pRetry from "p-retry";
import { createReadStream } from "node:fs";
import zlib from "node:zlib";
import { Transform, Writable } from "stream";
import { pipeline } from "stream/promises";
import { execa } from "execa";
import { rm } from "node:fs/promises";
import { Env } from "@humanwhocodes/env";

const env = new Env();

const { POLYGON_S3_ACCESS_KEY_ID, POLYGON_S3_SECRET_ACCESS_KEY } = env.required;

const sqliteDb = await open({
  filename: "/tmp/sync-state.db",
  driver: sqlite3.Database,
});
await sqliteDb.exec(`
  CREATE TABLE IF NOT EXISTS option_contract_sync_states (
    symbol TEXT,
    date TEXT,
    status TINYINT,
    UNIQUE (symbol, date) ON CONFLICT REPLACE
  )
`);
await sqliteDb.exec(`
  CREATE TABLE IF NOT EXISTS option_contract_aggregates_sync_states (
    ticker TEXT,
    status TINYINT,
    UNIQUE (ticker) ON CONFLICT REPLACE
  )
`);

export async function getPullOptionContractsState(
  symbol: string,
  date: string
) {
  const state = await sqliteDb.get(
    `
    SELECT
      *
    FROM option_contract_sync_states
    WHERE symbol = :symbol
    AND date = :date
  `,
    {
      ":symbol": symbol,
      ":date": date,
    }
  );
  return state;
}

const enum OptionContractSyncStatus {
  STARTED = 0,
  COMPLETED = 1,
}
type OptionContractSyncState = {
  status: OptionContractSyncStatus;
};
export async function setPullOptionContractsState(
  symbol: string,
  date: string,
  state: OptionContractSyncState
) {
  await sqliteDb.run(
    `
    INSERT INTO option_contract_sync_states (symbol, date, status) VALUES (:symbol, :date, :status)
  `,
    {
      ":symbol": symbol,
      ":date": date,
      ":status": state.status,
    }
  );
}

export async function getPullAggregatesState(ticker: string) {
  const state = await sqliteDb.get(
    `
    SELECT
      *
    FROM option_contract_aggregates_sync_states
    WHERE ticker = :ticker
  `,
    {
      ":ticker": ticker,
    }
  );
  return state;
}

const enum OptionContractAggregatesSyncStatus {
  STARTED = 0,
  COMPLETED = 1,
}
type OptionContractAggregatesSyncState = {
  status: OptionContractAggregatesSyncStatus;
};
export async function setPullAggregatesState(
  ticker: string,
  state: OptionContractAggregatesSyncState
) {
  await sqliteDb.run(
    `
    INSERT INTO option_contract_aggregates_sync_states (ticker, status) VALUES (:ticker, :status)
  `,
    {
      ":ticker": ticker,
      ":status": state.status,
    }
  );
}

export async function pullOptionContracts(symbol: string, date: string) {
  // check if sync was completed:
  if (
    (await getPullOptionContractsState(symbol, date))?.status !==
    OptionContractSyncStatus.COMPLETED
  ) {
    await setPullOptionContractsState(symbol, date, {
      status: OptionContractSyncStatus.STARTED,
    });
    for await (const batch of Polygon.makeGetOptionContractsIterator(
      symbol,
      date
    )) {
      console.log(batch.length);
      await pRetry(
        async () => {
          await clickhouse.insert({
            table: "option_contract_existences",
            values: batch,
            format: "JSONEachRow",
          });
        },
        { forever: true, factor: 2, maxTimeout: 120000 }
      );
    }
    await setPullOptionContractsState(symbol, date, {
      status: OptionContractSyncStatus.COMPLETED,
    });
  }
}

export async function pullOptionContractAggregates(
  optionContract: OptionContract
) {
  const ticker = Polygon.optionContractToTicker(optionContract);
  // check if sync was completed:
  if (
    (await getPullAggregatesState(ticker))?.status !==
    OptionContractAggregatesSyncStatus.COMPLETED
  ) {
    await setPullAggregatesState(ticker, {
      status: OptionContractAggregatesSyncStatus.STARTED,
    });
    const { firstDate } = await getOptionContractDateRange(optionContract);
    for await (const batch of Polygon.makeGetOptionContractAggregatesIterator({
      ...optionContract,
      firstDate,
    })) {
      if (batch.length > 0) {
        console.log(
          optionContract.symbol,
          optionContract.expirationDate,
          optionContract.strike,
          optionContract.type,
          new Date(batch[0].tsStart * 1000),
          new Date(batch[batch.length - 1].tsStart * 1000)
        );
        await pRetry(
          async () => {
            await clickhouse.insert({
              table: "option_contract_aggregates",
              values: batch,
              format: "JSONEachRow",
            });
          },
          { forever: true, factor: 2, maxTimeout: 120000 }
        );
      }
    }
    await setPullAggregatesState(ticker, {
      status: OptionContractAggregatesSyncStatus.COMPLETED,
    });
  }
}

export async function pullStockAggregates(
  symbol: string,
  firstDate: string,
  lastDate: string
) {
  // check if sync was completed:
  if (
    (await getPullAggregatesState(symbol))?.status !==
    OptionContractAggregatesSyncStatus.COMPLETED
  ) {
    await setPullAggregatesState(symbol, {
      status: OptionContractAggregatesSyncStatus.STARTED,
    });
    for await (const batch of Polygon.makeGetStockAggregatesIterator({
      symbol,
      firstDate,
      lastDate,
    })) {
      if (batch.length > 0) {
        console.log(
          symbol,
          new Date(batch[0].tsStart * 1000),
          new Date(batch[batch.length - 1].tsStart * 1000)
        );
        await clickhouse.insert({
          table: "stock_aggregates",
          values: batch,
          format: "JSONEachRow",
        });
        // await pRetry(
        //   async () => {
        //     console.log(`inserting ${batch.length} rows`);
        //     await clickhouse.insert({
        //       table: "stock_aggregates",
        //       values: batch,
        //       format: "JSONEachRow",
        //     });
        //     console.log("inserted");
        //   },
        //   { forever: true, factor: 2, maxTimeout: 120000 }
        // );
      }
    }
    await setPullAggregatesState(symbol, {
      status: OptionContractAggregatesSyncStatus.COMPLETED,
    });
  }
}

export async function pullAllStockAggregates(firstDate, lastDate) {
  const symbols = (
    await query(
      `SELECT DISTINCT(symbol) as symbol FROM option_contract_existences WHERE asOfDate >= '${firstDate}' AND asOfDate <= '${lastDate}'`
    )
  ).map(({ symbol }) => symbol);
  for (const symbol of symbols) {
    await pullStockAggregates(symbol, firstDate, lastDate);
  }
}

export async function* makeGetOptionContractsIterator(
  symbol: string,
  asOfDate: string
) {
  const limit = 2000;
  let offset = 0;
  let batch;
  do {
    batch = await query(`
      SELECT
        *
      FROM option_contract_existences
      WHERE asOfDate = '${asOfDate}' 
      AND symbol = '${symbol}'
      ORDER BY expirationDate ASC, strike ASC, type ASC
      LIMIT ${limit}
      OFFSET ${offset}
    `);
    yield batch;
    offset = offset + limit;
  } while (batch.length === limit);
}

export async function pullOptionContractsSince(
  symbol: string,
  firstDate: string
) {
  const currentDateAsDateObject = new Date(firstDate);
  const yesterdayAsDateObject = new Date();
  yesterdayAsDateObject.setUTCDate(yesterdayAsDateObject.getUTCDate() - 1);
  while (currentDateAsDateObject <= yesterdayAsDateObject) {
    const currentDate = currentDateAsDateObject.toISOString().substring(0, 10);
    console.log(`Date: ${currentDate}:`);
    await pullOptionContracts(symbol, currentDate);
    currentDateAsDateObject.setUTCDate(
      currentDateAsDateObject.getUTCDate() + 1
    );
  }
}

export async function pullOptionContractAggregatesSince(
  symbol: string,
  firstDate: string
) {
  const currentDateAsDateObject = new Date(firstDate);
  const yesterdayAsDateObject = new Date();
  yesterdayAsDateObject.setUTCDate(yesterdayAsDateObject.getUTCDate() - 1);
  while (currentDateAsDateObject <= yesterdayAsDateObject) {
    const currentDate = currentDateAsDateObject.toISOString().substring(0, 10);
    console.log(`Date: ${currentDate}`);
    for await (const optionContracts of makeGetOptionContractsIterator(
      symbol,
      currentDate
    )) {
      for (const optionContract of optionContracts) {
        await pullOptionContractAggregates(optionContract);
      }
    }
    currentDateAsDateObject.setUTCDate(
      currentDateAsDateObject.getUTCDate() + 1
    );
  }
}

export async function getOptionContractDateRange({
  symbol,
  expirationDate,
  strike,
  type,
}: OptionContract) {
  const rows = await query<{ firstDate: string; lastDate: string }>(`
    SELECT
      min(asOfDate) AS firstDate,
      max(asOfDate) AS lastDate
    FROM option_contract_existences
    WHERE symbol = '${symbol}'
    AND expirationDate = '${expirationDate}'
    AND strike = ${strike}
    AND type = '${type}'
  `);
  return rows[0] || { firstDate: null, lastDate: null };
}

function transformCsvLineToObject(line: string) {
  const [
    ticker,
    volume,
    open,
    close,
    high,
    low,
    window_start,
    numberOfTransactions,
  ] = line.split(",");
  const symbol = ticker.substring(2, ticker.length - 15);
  const tickerDate = ticker.substring(ticker.length - 15, ticker.length - 9);
  const expirationDate = `20${tickerDate.substring(
    0,
    2
  )}-${tickerDate.substring(2, 4)}-${tickerDate.substring(4)}`;
  const type =
    ticker.substring(ticker.length - 9, ticker.length - 8) === "C"
      ? "call"
      : "put";
  const strike = parseInt(ticker.substring(ticker.length - 8)) / 1000;
  /** UNIX time in seconds */
  const tsStart = parseInt(window_start.substring(0, window_start.length - 9));
  return {
    symbol,
    expirationDate,
    strike,
    type,
    tsStart,
    open,
    close,
    low,
    high,
    volume,
    volumeWeightedPrice: 0,
  };
}

export async function uploadCsvToClickhouse(filename: string) {
  let buf = "";
  for await (const chunk of createReadStream(filename, {
    start: 60 /* skip header */,
    highWaterMark: 1024 * 1024,
  })) {
    const lines = buf.concat(chunk).split(/\r?\n/);
    buf = lines.pop() ?? "";

    await clickhouse.insert({
      table: "option_contract_aggregates",
      values: lines.map(transformCsvLineToObject),
      format: "JSONEachRow",
    });
  }
  if (buf.length) {
    // last line, if file does not end with newline
    await clickhouse.insert({
      table: "option_contract_aggregates",
      values: [transformCsvLineToObject(buf)],
      format: "JSONEachRow",
    });
  }
}

const accessKeyId = POLYGON_S3_ACCESS_KEY_ID;
const secretAccessKey = POLYGON_S3_SECRET_ACCESS_KEY;
// const X_Amz_Algorithm = "AWS4-HMAC-SHA256";
// const X_Amz_Credential =
//   "bfe011b0-01e7-4c16-aedf-52cb83722c36/20240702/us-east-1/s3/aws4_request";
// const X_Amz_Expires = "900";
// const X_Amz_SignedHeaders = "host";
// const X_Amz_Signature =
//   "e9fd0ac569c4d5fd5757ba4104e95ce4cfd0c17b16648ec32f978265d4188f37";
export async function ingestOptionContractAggregateFlatfile(date: string) {
  let buf = "";
  let skippedFirstLine = false;
  const [year, month, day] = date.split("-");
  const localFilename = `/tmp/${date}.csv.gz`;
  try {
    await execa({
      env: {
        AWS_ACCESS_KEY_ID: accessKeyId,
        AWS_SECRET_ACCESS_KEY: secretAccessKey,
        S3_ENDPOINT_URL: "https://files.polygon.io",
        AWS_REGION: "us-east-1",
      },
    })("s5cmd", [
      "cp",
      `s3://flatfiles/us_options_opra/minute_aggs_v1/${year}/${month}/${date}.csv.gz`,
      localFilename,
    ]);
  } catch (err) {
    return;
    // if (err.includes("status code: 404")) {
    //   return;
    // } else {
    //   console.error("error downloading flatfile from polygon s3", err);
    //   throw err;
    // }
  }

  try {
    await pipeline(
      createReadStream(localFilename, {
        highWaterMark: 1024 * 1024,
      }),
      zlib.createGunzip(),
      new Transform({
        transform(chunk, encoding, next) {
          const lines = buf.concat(chunk).split(/\r?\n/);
          if (!skippedFirstLine) {
            lines.shift();
            skippedFirstLine = true;
          }
          buf = lines.pop();
          next(null, lines);
        },
        objectMode: true,
      }),
      new Writable({
        objectMode: true,
        async write(lines, encoding, next) {
          // console.log(lines.map(transformCsvLineToObject));
          await clickhouse.insert({
            table: "option_contract_aggregates",
            values: lines.map(transformCsvLineToObject),
            format: "JSONEachRow",
          });
          next();
        },
      })
    );
  } catch (err) {
    console.error(err);
  }
  await rm(localFilename);
  console.log("done");
}

export async function pullOptionContractAggregatesFromFlatFileSince(
  firstDate: string,
  lastDate?: string
) {
  const currentDateAsDateObject = new Date(firstDate);
  const yesterdayAsDateObject = new Date();
  yesterdayAsDateObject.setUTCDate(yesterdayAsDateObject.getUTCDate() - 1);
  const lastDateAsDateObject = lastDate
    ? new Date(lastDate)
    : yesterdayAsDateObject;
  while (currentDateAsDateObject <= lastDateAsDateObject) {
    const currentDate = currentDateAsDateObject.toISOString().substring(0, 10);
    console.log(`Date: ${currentDate}`);
    await ingestOptionContractAggregateFlatfile(currentDate);
    currentDateAsDateObject.setUTCDate(
      currentDateAsDateObject.getUTCDate() + 1
    );
  }
}
